package org.nkumar.ant.trace.locator;

import org.nkumar.ant.trace.action.Logger;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * This is the main factory class to load information about xml content within a build file.
 * Basically the xml log of a build file spits out information like the build file name,
 * the target or the task name and starting line number.(actually it is the number of the line
 * where the starting element of the tag ends). From this info the xml content corresponding to
 * the tag has to fetched so that it can be shown against xml log.
 */
public final class ElementLocatorFactory
{
    private final Map fileContentMap = new HashMap();

    private final Map fileRawContentMap = new HashMap();

    public ElementLocatorFactory()
    {
    }

    public ElementContent getContentForElement(final String elementName, final String fileName, final int lineNumber)
            throws IOException, ParserConfigurationException, SAXException
    {
        return (ElementContent) getContentMap(fileName)
                .get(new ElementNameStartLocationKey(elementName, lineNumber));
    }

    private Map getContentMap(final String fileName)
            throws IOException, ParserConfigurationException, SAXException
    {
        Logger.debug("ElementLocatorFactory.getContentMap");
        final String canonicalFileName = getCanonicalFileName(fileName);
        if (!fileContentMap.containsKey(canonicalFileName))
        {
            fileContentMap.put(fileName, createContentMapForFile(fileName));
        }
        return (Map) fileContentMap.get(canonicalFileName);
    }

    private Map createContentMapForFile(final String fileName)
            throws ParserConfigurationException, SAXException, IOException
    {
        Logger.debug("ElementLocatorFactory.createContentMapForFile");
        final Map map = new HashMap();
        final DefaultHandler handler = new LocatorContentHandler();
        final SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(true);
        factory.newSAXParser().parse(new File(fileName), handler);
        final List elementList = ((LocatorContentHandler) handler).getElementList();
        for (Iterator itr = elementList.iterator(); itr.hasNext();)
        {
            final ElementLocation location = (ElementLocation) itr.next();
            final ElementNameStartLocationKey key = new ElementNameStartLocationKey(location.getElementName(),
                    location.getStartElementEndLocator().getLineNumber());
            map.put(key, new ElementContent(location, this));
        }
        Logger.debug("map = " + map);
        return map;

    }

    String[] getFileContent(final String fileName, final int startLineNumber, final int endLineNumber)
            throws IOException
    {
        if (endLineNumber < startLineNumber)
        {
            throw new IllegalArgumentException("ending (" + endLineNumber
                    + ") before starting (" + startLineNumber + ")");
        }
        final String canonicalFileName = getCanonicalFileName(fileName);
        if (!fileRawContentMap.containsKey(canonicalFileName))
        {
            final BufferedReader in = new BufferedReader(new InputStreamReader(
                    new FileInputStream(canonicalFileName), "UTF-8"));
            final List lines = new ArrayList();
            String line;
            while ((line = in.readLine()) != null)
            {
                lines.add(line);
            }
            in.close();
            fileRawContentMap.put(canonicalFileName, lines.toArray(new String[lines.size()]));
        }
        final String[] fileContent = (String[]) fileRawContentMap.get(canonicalFileName);
        final int startingIndex = startLineNumber - 1;     //xml parsers numbers start with 1
        final int length = Math.min(fileContent.length, endLineNumber) - startingIndex;
        if (length == -1)
        {
            //highly unlikely
            return new String[0];
        }
        final String[] elementContent = new String[length];
        System.arraycopy(fileContent, startingIndex, elementContent, 0, length);
        return elementContent;
    }

    private static String getCanonicalFileName(final String fileName)
    {
        try
        {
            return new File(fileName).getCanonicalPath();
        }
        catch (IOException e)
        {
            throw new RuntimeException("Unable to canonicalize the file path : " + fileName, e);
        }
    }
}
