package org.nkumar.ant.trace.locator;

import org.xml.sax.Locator;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * An immutable impl of the locator interface.
 * Used to store the current content of the content handler locator for future use.
 */
final class ImmutableLocator implements Locator
{
    private final int lineNumber;
    private final int columnNumber;
    private final String fileName;

    public ImmutableLocator(final Locator locator)
    {
        lineNumber = locator.getLineNumber();
        columnNumber = locator.getColumnNumber();
        fileName = systemIdToFileName(locator).intern();
    }

    private static String systemIdToFileName(final Locator locator)
    {
        final String systemId = locator.getSystemId();
        try
        {
            final File file = new File(fromURI(systemId));
            return file.getCanonicalPath();
        }
        catch (Exception e)
        {
            throw new RuntimeException("Unable to locate the file : " + systemId, e);
        }
    }

    //lifted from ant code
    private static String fromURI(String uri)
    {
        URL url = null;
        try
        {
            url = new URL(uri);
        }
        catch (MalformedURLException ignore)
        {
            //exception thrown later
        }
        if (url == null || !("file".equals(url.getProtocol())))
        {
            throw new IllegalArgumentException("Can only handle valid file: URIs");
        }
        final StringBuffer buf = new StringBuffer(url.getHost());
        if (buf.length() > 0)
        {
            buf.insert(0, File.separatorChar).insert(0, File.separatorChar);
        }

        final String file = url.getFile();
        final int queryPos = file.indexOf('?');
        buf.append((queryPos < 0) ? file : file.substring(0, queryPos));

        uri = buf.toString().replace('/', File.separatorChar);

        if (File.pathSeparatorChar == ';' && uri.startsWith("\\") && uri.length() > 2
                && Character.isLetter(uri.charAt(1)) && uri.lastIndexOf(':') > -1)
        {
            uri = uri.substring(1);
        }
        return uri;
    }

    public int getColumnNumber()
    {
        return columnNumber;
    }

    public int getLineNumber()
    {
        return lineNumber;
    }

    public String getPublicId()
    {
        throw new UnsupportedOperationException("TODO");
    }

    public String getSystemId()
    {
        return fileName;
    }

    public String toString()
    {
        return fileName + "@" + lineNumber + "," + columnNumber;
    }

}
