package org.nkumar.ant.trace.locator;

import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.helpers.DefaultHandler;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Stack;

/**
 * this handler keeps track of where any element started and ended.
 * The locator passed while parsing point to location just
 * where the point is after the event being handled.
 * So if you in beginelement then the location point to the end of the start element.
 * Hence you have to keep track of the previous location to find the location
 * of the start of the element.
 */
final class LocatorContentHandler extends DefaultHandler
{
    //reference to the location where the last event ended
    //when you get a callback, the locator points to the location where the event ended
    //for startElement the locator will point to the end of the start element tag.
    //this can be used as the starting point of the next event.
    //this would be an instance of immutable locator
    private Locator lastLocator;

    // the locator that is passed from the parser
    private Locator parserLocator;

    //Stack<ElementLocation>
    //holds element location of the current elements and its ancestors
    private Stack elementStack = new Stack();

    //List<ElementLocation>
    //holds element locations of all the elements parsed yet
    private List elementList = new ArrayList();

    //stores the reference of the locator passed from the parser
    //this method may not be called if the parser does not support locator
    //in that case this class won't work
    public void setDocumentLocator(final Locator locator)
    {
        this.parserLocator = locator;
        lastLocator = new ImmutableLocator(parserLocator);
    }

    //see the details about lastLocator
    private void setLastLocator()
    {
        if (parserLocator != null)
        {
            lastLocator = new ImmutableLocator(parserLocator);
        }
        else
        {
            throw new RuntimeException("Locator not supported by the parser being used");
        }
    }

    public void startElement(final String uri, final String localName, final String qName, final Attributes attributes)
    {
        final Locator startLocator = lastLocator;
        setLastLocator();
        final ElementLocation item = new ElementLocation(qName, startLocator, lastLocator);
        elementStack.push(item);
        elementList.add(item);
    }

    public void endElement(final String uri, final String localName, final String qName)
    {
        final Locator startLocator = lastLocator;
        setLastLocator();
        final ElementLocation elementLocation = (ElementLocation) elementStack.pop();
        elementLocation.setEndElementLocators(startLocator, lastLocator);
    }

    public void characters(final char[] ch, final int start, final int length)
    {
        setLastLocator();
    }

    public void processingInstruction(final String target, final String data)
    {
        setLastLocator();
    }

    public void ignorableWhitespace(final char[] ch, final int start, final int length)
    {
        setLastLocator();
    }

    public List getElementList()
    {
        return Collections.unmodifiableList(elementList);
    }
}
