package org.nkumar.ant.trace.logger;

import org.apache.tools.ant.BuildEvent;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.BuildLogger;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Target;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.UnknownElement;
import org.apache.tools.ant.util.DOMElementWriter;
import org.apache.tools.ant.util.DateUtils;
import org.apache.tools.ant.util.StringUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.Writer;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

/**
 * Generates a file in the current directory with
 * an XML description of what happened during a build.
 * The default filename is "log.xml", but this can be overridden
 * with the property <code>XmlLogger.file</code>.
 *
 * This implementation assumes in its sanity checking that only one
 * thread runs a particular target/task at a time. This is enforced
 * by the way that parallel builds and antcalls are done - and
 * indeed all but the simplest of tasks could run into problems
 * if executed in parallel.
 * @see Project#addBuildListener(org.apache.tools.ant.BuildListener)
 */
//ant -listener org.nkumar.ant.trace.logger.XmlLogger
//TODO donot depend on ant to give you the correct parent
//use the stack information for the same
public final class XmlLogger implements BuildLogger
{
    private int msgOutputLevel = Project.MSG_DEBUG;
    private PrintStream outStream;
    private PrintStream debugWriter;

    /**
     * DocumentBuilder to use when creating the document to start with.
     */
    private static DocumentBuilder builder = getDocumentBuilder();

    /**
     * Returns a default DocumentBuilder instance or throws an
     * ExceptionInInitializerError if it can't be created.
     * @return a default DocumentBuilder instance.
     */
    private static DocumentBuilder getDocumentBuilder()
    {
        try
        {
            return DocumentBuilderFactory.newInstance().newDocumentBuilder();
        }
        catch (Exception exc)
        {
            throw new ExceptionInInitializerError(exc);
        }
    }

    /**
     * The complete log document for this build.
     */
    private Document doc = builder.newDocument();
    /**
     * Mapping for when tasks started (Task to TimedElement).
     */
    private Map tasks = Collections.synchronizedMap(new HashMap());
    /**
     * Mapping for when targets started (Task to TimedElement).
     */
    private Map targets = Collections.synchronizedMap(new HashMap());
    /**
     * When the build started.
     */
    private TimedElement buildElement;

    /**
     * Utility class representing the time an element started.
     */
    private static final class TimedElement
    {
        /**
         * Start time in milliseconds
         * (as returned by <code>System.currentTimeMillis()</code>).
         */
        private long startTime;
        /**
         * Element created at the start time.
         */
        private Element element;

        public String toString()
        {
            return element.getTagName() + ":" + element.getAttribute("name");
        }
    }

    /**
     * Constructs a new BuildListener that logs build events to an XML file.
     */
    public XmlLogger()
    {
//        createDebugWriter();
    }

    private void createDebugWriter()
    {
        try
        {
            debugWriter = new PrintStream(new FileOutputStream("debug.log"));
        }
        catch (IOException e)
        {
            throw new RuntimeException(e);
        }
    }

    private void debug(String str)
    {
        if (debugWriter != null)
        {
            debugWriter.println(str);
        }
    }

    /**
     * Fired when the build starts, this builds the top-level element for the
     * document and remembers the time of the start of the build.
     * @param event Ignored.
     */
    public void buildStarted(final BuildEvent event)
    {
        buildElement = new TimedElement();
        buildElement.startTime = System.currentTimeMillis();
        buildElement.element = doc.createElement("build");
        debug("build started");
    }

    /**
     * Fired when the build finishes, this adds the time taken and any
     * error stacktrace to the build element and writes the document to disk.
     * @param event An event with any relevant extra information.
     * Will not be <code>null</code>.
     */
    public void buildFinished(final BuildEvent event)
    {
        debug("build finished");
        if (debugWriter != null)
        debugWriter.close();
        final long totalTime = System.currentTimeMillis() - buildElement.startTime;
        buildElement.element.setAttribute("time",
                DateUtils.formatElapsedTime(totalTime));

        if (event.getException() != null)
        {
            buildElement.element.setAttribute("error",
                    event.getException().toString());
            // print the stacktrace in the build file it is always useful...
            // better have too much info than not enough.
            final Throwable t = event.getException();
            final Text errText = doc.createCDATASection(StringUtils.getStackTrace(t));
            final Element stacktrace = doc.createElement("stacktrace");
            stacktrace.appendChild(errText);
            buildElement.element.appendChild(stacktrace);
        }

        String outFilename = event.getProject().getProperty("XmlLogger.file");
        if (outFilename == null)
        {
            outFilename = "log.xml";
        }
        String xslUri
                = event.getProject().getProperty("ant.XmlLogger.stylesheet.uri");
        if (xslUri == null)
        {
            xslUri = "log.xsl";
        }
        Writer out = null;
        try
        {
            // specify output in UTF8 otherwise accented characters will blow
            // up everything
            OutputStream stream = outStream;
            if (stream == null)
            {
                stream = new FileOutputStream(outFilename);
            }
            out = new OutputStreamWriter(stream, "UTF8");
            out.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
            if (xslUri.length() > 0)
            {
                out.write("<?xml-stylesheet type=\"text/xsl\" href=\""
                        + xslUri + "\"?>\n\n");
            }
            (new DOMElementWriter()).write(buildElement.element, out, 0, "\t");
            out.flush();
        }
        catch (IOException exc)
        {
            throw new BuildException("Unable to write log file", exc);
        }
        finally
        {
            if (out != null)
            {
                try
                {
                    out.close();
                }
                catch (IOException ignore)
                {
                    // ignore
                }
            }
        }
        buildElement = null;
    }

    private static ThreadLocal localStack = new ThreadLocal()
    {
        protected synchronized Object initialValue()
        {
            return new Stack();
        }
    };

    /**
     * Returns the stack of timed elements for the current thread.
     * @return the stack of timed elements for the current thread
     */
    private static Stack getStack()
    {
        return (Stack) localStack.get();
    }

    /**
     * Fired when a target starts building, this pushes a timed element
     * for the target onto the stack of elements for the current thread,
     * remembering the current time and the name of the target.
     * @param event An event with any relevant extra information.
     * Will not be <code>null</code>.
     */
    public void targetStarted(final BuildEvent event)
    {
        final Target target = event.getTarget();
        final TimedElement targetElement = new TimedElement();
        targetElement.startTime = System.currentTimeMillis();
        targetElement.element = doc.createElement("target");
        targetElement.element.setAttribute("name", target.getName());
        targetElement.element.setAttribute("location", target.getLocation().toString());
        targets.put(target, targetElement);
        getStack().push(targetElement);
        debug("target started name=" + target.getName() + " loaction=" + target.getLocation());
    }

    /**
     * Fired when a target finishes building, this adds the time taken
     * and any error stacktrace to the appropriate target element in the log.
     * @param event An event with any relevant extra information.
     * Will not be <code>null</code>.
     */
    public void targetFinished(final BuildEvent event)
    {
        final Target target = event.getTarget();
        final TimedElement targetElement = (TimedElement) targets.get(target);
        if (targetElement != null)
        {
            debug("target finished name=" + target.getName() + " loaction=" + target.getLocation());
            final long totalTime
                    = System.currentTimeMillis() - targetElement.startTime;
            targetElement.element.setAttribute("time",
                    DateUtils.formatElapsedTime(totalTime));

            TimedElement parentElement = null;
            final Stack threadStack = getStack();
            if (!threadStack.empty())
            {
                final TimedElement poppedStack = (TimedElement) threadStack.pop();
                if (poppedStack != targetElement)
                {
                    throw new RuntimeException("Mismatch - popped element = "
                            + poppedStack
                            + " finished target element = "
                            + targetElement);
                }
                if (!threadStack.empty())
                {
                    parentElement = (TimedElement) threadStack.peek();
                }
            }
            if (parentElement == null)
            {
                buildElement.element.appendChild(targetElement.element);
            }
            else
            {
                parentElement.element.appendChild(targetElement.element);
            }
        }
        targets.remove(target);
    }

    /**
     * Fired when a task starts building, this pushes a timed element
     * for the task onto the stack of elements for the current thread,
     * remembering the current time and the name of the task.
     * @param event An event with any relevant extra information.
     * Will not be <code>null</code>.
     */
    public void taskStarted(final BuildEvent event)
    {
        final TimedElement taskElement = new TimedElement();
        taskElement.startTime = System.currentTimeMillis();
        taskElement.element = doc.createElement("task");

        final Task task = event.getTask();
        String name = event.getTask().getTaskName();
        if (name == null)
        {
            name = "";
        }
        taskElement.element.setAttribute("name", name);
        taskElement.element.setAttribute("location",
                task.getLocation().toString());
        debug("task started name=" + task.getTaskName() + " loaction=" + task.getLocation());
        tasks.put(task, taskElement);
        getStack().push(taskElement);
    }

    /**
     * Fired when a task finishes building, this adds the time taken
     * and any error stacktrace to the appropriate task element in the log.
     * @param event An event with any relevant extra information.
     * Will not be <code>null</code>.
     */
    public void taskFinished(final BuildEvent event)
    {
        final Task task = event.getTask();
        final TimedElement taskElement = (TimedElement) tasks.get(task);
        if (taskElement != null)
        {
            debug("task finished name=" + task.getTaskName() + " loaction=" + task.getLocation());
            final long totalTime = System.currentTimeMillis() - taskElement.startTime;
            taskElement.element.setAttribute("time",
                    DateUtils.formatElapsedTime(totalTime));
            final Stack threadStack = getStack();
            if (!threadStack.empty())
            {
                final TimedElement poppedStack = (TimedElement) threadStack.pop();
                if (poppedStack != taskElement)
                {
                    throw new RuntimeException("Mismatch - popped element = "
                            + poppedStack + " finished task element = "
                            + taskElement);
                }
            }
            tasks.remove(task);
            if (!threadStack.empty())
            {
                final TimedElement parent = (TimedElement) threadStack.peek();
                parent.element.appendChild(taskElement.element);
            }
        }
        else
        {
            throw new RuntimeException("Unknown task " + task + " not in " + tasks);
        }
    }


    /**
     * Get the TimedElement associated with a task.
     *
     * Where the task is not found directly, search for unknown elements which
     * may be hiding the real task
     */
    private TimedElement getTaskElement(final Task task)
    {
        final TimedElement element = (TimedElement) tasks.get(task);
        if (element != null)
        {
            return element;
        }

        for (Iterator e = tasks.keySet().iterator(); e.hasNext();)
        {
            final Task key = (Task) e.next();
            if (key instanceof UnknownElement)
            {
                if (((UnknownElement) key).getTask() == task)
                {
                    return (TimedElement) tasks.get(key);
                }
            }
        }

        return null;
    }

    /**
     * Fired when a message is logged, this adds a message element to the
     * most appropriate parent element (task, target or build) and records
     * the priority and text of the message.
     * @param event An event with any relevant extra information.
     * Will not be <code>null</code>.
     */
    public void messageLogged(final BuildEvent event)
    {
        final int priority = event.getPriority();
        if (priority > msgOutputLevel)
        {
            return;
        }
        final Element messageElement = doc.createElement("message");

        final String name = getPriorityAsString(priority);
        messageElement.setAttribute("priority", name);

        final Text messageText = doc.createTextNode(event.getMessage());
        messageElement.appendChild(messageText);

        final Stack stack = getStack();
        TimedElement parentElement = stack.isEmpty() ? null : (TimedElement) stack.peek();

//        Task task = event.getTask();
//
//        if (task != null)
//        {
//            parentElement = getTaskElement(task);
//        }
//        Target target = event.getTarget();
//        if (parentElement == null && target != null)
//        {
//            parentElement = (TimedElement) targets.get(target);
//        }
        if (parentElement == null)
        {
            parentElement = buildElement;
        }
        parentElement.element.appendChild(messageElement);
        debug("message at " + name + ":" + messageText);
    }

    private static String getPriorityAsString(final int priority)
    {
        final String name;
        switch (priority)
        {
            case Project.MSG_ERR:
                name = "error";
                break;
            case Project.MSG_WARN:
                name = "warn";
                break;
            case Project.MSG_INFO:
                name = "info";
                break;
            default:
                name = "debug";
                break;
        }
        return name;
    }

    // -------------------------------------------------- BuildLogger interface

    /**
     * Set the logging level when using this as a Logger
     * @param level the logging level -
     * see {@link org.apache.tools.ant.Project#MSG_ERR Project}
     * class for level definitions
     */
    public void setMessageOutputLevel(final int level)
    {
        msgOutputLevel = level;
    }

    /**
     * Set the output stream to which logging output is sent when operating
     * as a logger.
     * @param output the output PrintStream.
     */
    public void setOutputPrintStream(final PrintStream output)
    {
        this.outStream = new PrintStream(output, true);
    }

    /**
     * Ignore emacs mode, as it has no meaning in XML format
     * @param emacsMode true if logger should produce emacs compatible
     * output
     */
    public void setEmacsMode(final boolean emacsMode)
    {
    }

    /**
     * Ignore error print stream. All output will be written to
     * either the XML log file or the PrintStream provided to
     * setOutputPrintStream
     * @param err the stream we are going to ignore.
     */
    public void setErrorPrintStream(final PrintStream err)
    {
    }

}
