package org.nkumar.ant.trace.model;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.nkumar.ant.trace.action.Logger;
import org.nkumar.ant.trace.locator.ElementContent;
import org.nkumar.ant.trace.locator.ElementLocatorFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public final class DataHolderBuilder
{
    private final List mappingList = new ArrayList();
    private final ElementLocatorFactory elf = new ElementLocatorFactory();

    public DataHolderBuilder()
    {
    }

    /**
     * Given the xml log as string try and build the dataholder object.
     * Converts the xml log on the to a textual log and that
     * more or less becomes the content of the left pane.
     * The xml log has location values for tasks and targets.
     * This is used to load the corresponding xml content from the build file.
     * These are then mapped to the content in the log.
     * There are quite a few complexities here like tags being nested.
     * ant and antcall tasks.
     * There might be somethings more like parallel and ant import
     * that has not been handled yet.
     */
    public DataHolder createDataHolder(final String xmlLog) throws IOException, JDOMException
    {
        Logger.debug("DataHolderBuilder.createDataHolder");
        final Document docLog = new SAXBuilder().build(new StringReader(xmlLog));
        final Element root = docLog.getRootElement();
        if (!"build".equals(root.getName()))
        {
            return null;
        }
        visitBuild(root);
        cleanupMappingList(mappingList);
        return createDataHolderFromMappingList(mappingList);
    }

    private static DataHolder createDataHolderFromMappingList(final List list)
    {
        Logger.debug("DataHolderBuilder.createDataHolderFromMappingList");
        final StringBuffer left = new StringBuffer(10240);
        final StringBuffer right = new StringBuffer(10240);
        final List leftLineNumbers = new ArrayList();
        final List rightLineNumbers = new ArrayList();
        int leftCount = 0;
        int rightCount = 0;
        leftLineNumbers.add(new Integer(leftCount));
        rightLineNumbers.add(new Integer(rightCount));
        for (int i = 0; i < list.size(); i++)
        {
            final Mappings mappings = (Mappings) list.get(i);
            left.append(mappings.getLeftContent());
            right.append(mappings.getRightContent());
            leftCount += mappings.getLeftContentSize();
            rightCount += mappings.getRightContentSize();
            leftLineNumbers.add(new Integer(leftCount));
            rightLineNumbers.add(new Integer(rightCount));
        }
        Logger.debug("left = " + left);
        Logger.debug("right = " + right);
        Logger.debug("leftLineNumbers = " + leftLineNumbers);
        Logger.debug("rightLineNumbers = " + rightLineNumbers);
        return new DataHolder(left.toString(), getLineNumbers(leftLineNumbers),
                right.toString(), getLineNumbers(rightLineNumbers));
    }

    private static int[] getLineNumbers(final List list)
    {
        Logger.debug("DataHolderBuilder.getLineNumbers");
        final int[] lineNumbers = new int[list.size()];
        for (int i = 0; i < list.size(); i++)
        {
            final Integer number = (Integer) list.get(i);
            lineNumbers[i] = number.intValue();
        }
        return lineNumbers;
    }

    private static void cleanupMappingList(final List mappingList)
    {
        Logger.debug("DataHolderBuilder.cleanupMappingList");
        for (Iterator iterator = mappingList.iterator(); iterator.hasNext();)
        {
            final Mappings mappings = (Mappings) iterator.next();
            if (mappings == null || mappings.isEmpty())
            {
                iterator.remove();
                continue;
            }
            mappings.fillUp();
        }
    }

    //TODO all these visit methods do not actually follow the visitor pattern
    //TODO refactor the visit code to remove duplicates

    private void visitBuild(final Element build)
    {
        Logger.debug("DataHolderBuilder.visitBuild");
        Mappings currentMappings = null;
        final List children = build.getChildren();
        for (int i = 0; i < children.size(); i++)
        {
            final Element child = (Element) children.get(i);
            if ("message".equals(child.getName()))
            {
                if (currentMappings == null)
                {
                    currentMappings = new Mappings();
                    mappingList.add(currentMappings);
                }
                visitMessage(child, currentMappings);
            }
            else
            {
                currentMappings = null;
                if ("target".equals(child.getName()))
                {
                    visitTarget(child);
                }
                else if ("task".equals(child.getName()))
                {
                    final Mappings mappings = new Mappings();
                    mappings.appendRightContent(child.getAttributeValue("location", ""));
                    visitTask(child, mappings);
                }
            }
        }
    }

    private void visitTarget(final Element target)
    {
        Logger.debug("DataHolderBuilder.visitTarget");
        final List children = target.getChildren();
        final ElementContent content = getElementContent(target, "target");
        if (content == null)
        {
            Logger.debug("target content == null");
            return;
        }
        final int startingMappingIndex = mappingList.size();
        final String[] startingContent = content.getStartingContent();
        final String[] endingContent = content.getEndingContent();

        for (int i = 0; i < children.size(); i++)
        {
            final Element child = (Element) children.get(i);
            if ("task".equals(child.getName()))
            {
                visitTask(child);
            }
            //TODO else
        }
        if (mappingList.size() > startingMappingIndex)
        {
            final Mappings startMapping = (Mappings) mappingList.get(startingMappingIndex);
            for (int i = startingContent.length - 1; i >= 0; i--)
            {
                startMapping.prependRightContent(startingContent[i]);
            }
            startMapping.prependRightContent(target.getAttributeValue("location", ""));
            if (endingContent.length > 0)
            {
                final Mappings endMapping = (Mappings) mappingList.get(mappingList.size() - 1);
                for (int i = 0; i < endingContent.length; i++)
                {
                    endMapping.appendRightContent(endingContent[i]);
                }
            }
        }
    }

    private ElementContent getElementContent(final Element element, String name)
    {
        Logger.debug("DataHolderBuilder.getElementContent");
        try
        {
            if (name == null)
            {
                name = element.getAttributeValue("name");
            }
            Logger.debug("name = " + name);
            String location = element.getAttributeValue("location");
            Logger.debug("location = " + location);
            int lastIndex = location.lastIndexOf(":");
            location = location.substring(0, lastIndex);
            Logger.debug("location = " + location);
            lastIndex = location.lastIndexOf(":");
            final int lineNumber = Integer.parseInt(location.substring(lastIndex + 1));
            Logger.debug("lineNumber = " + lineNumber);
            final String fileName = location.substring(0, lastIndex);
            Logger.debug("fileName = " + fileName);
            return elf.getContentForElement(name, fileName, lineNumber);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            return null;
        }
    }

    private void visitTask(final Element task)
    {
        visitTask(task, new Mappings());
    }

    private void visitTask(final Element task, final Mappings mappings)
    {
        Logger.debug("DataHolderBuilder.visitTask");
        final String taskName = task.getAttributeValue("name");
        if ("ant".equals(taskName) || "antcall".equals(taskName))
        {
            visitAntOrAntCallTask(task);
        }
        else
        {
            final ElementContent content = getElementContent(task, null);
            if (content == null)
            {
                Logger.debug("element content == null");
                return;
            }
            final List children = task.getChildren();
            if (!children.isEmpty())
            {
                mappingList.add(mappings);
            }
            final String[] strs = content.getContent();
            for (int i = 0; i < strs.length; i++)
            {
                mappings.appendRightContent(strs[i]);
            }
            for (int i = 0; i < children.size(); i++)
            {
                final Element child = (Element) children.get(i);
                if ("message".equals(child.getName()))
                {
                    visitMessage(child, mappings);
                }
                else
                {
                    //TODO not expected: should some error be thrown here
                }
            }
        }
    }

    private void visitAntOrAntCallTask(final Element task)
    {
        Logger.debug("DataHolderBuilder.visitAntOrAntCallTask");
        Mappings currentMappings = null;
        final List children = task.getChildren();
        final ElementContent content = getElementContent(task, null);
        if (content == null)
        {
            Logger.debug("task content == null for ant call");
            return;
        }
        final int startingMappingIndex = mappingList.size();
        final String[] startingContent = content.getStartingContent();
        final String[] endingContent = content.getEndingContent();

        for (int i = 0; i < children.size(); i++)
        {
            final Element child = (Element) children.get(i);
            if ("message".equals(child.getName()))
            {
                if (currentMappings == null)
                {
                    currentMappings = new Mappings();
                    mappingList.add(currentMappings);
                }
                visitMessage(child, currentMappings);
            }
            else
            {
                currentMappings = null;
                if ("target".equals(child.getName()))
                {
                    visitTarget(child);
                }
                else if ("task".equals(child.getName()))
                {
                    final Mappings mappings = new Mappings();
                    mappings.appendRightContent(child.getAttributeValue("location", ""));
                    visitTask(child, mappings);
                }
            }
        }

        if (mappingList.size() > startingMappingIndex)
        {
            final Mappings startMapping = (Mappings) mappingList.get(startingMappingIndex);
            for (int i = startingContent.length - 1; i >= 0; i--)
            {
                startMapping.prependRightContent(startingContent[i]);
            }
            if (endingContent.length > 0)
            {
                final Mappings endMapping = (Mappings) mappingList.get(mappingList.size() - 1);
                for (int i = 0; i < endingContent.length; i++)
                {
                    endMapping.appendRightContent(endingContent[i]);
                }
            }
        }
    }

    private static void visitMessage(final Element message, final Mappings mappings)
    {
        if (!"message".equals(message.getName()))
        {
            throw new IllegalArgumentException("expecting a message element instead of "
                    + message.getName());
        }
        final String msgText = message.getText();
        String priority = message.getAttributeValue("priority", "");
        priority = "[" + priority + "] ";
        mappings.appendLeftContent(getPadding(10 - priority.length()) + priority + msgText);
    }

    private static String getPadding(final int length)
    {
        if (length < 1)
        {
            return "";
        }
        switch (length)
        {
            case 1:
                return " ";
            case 2:
                return "  ";
            case 3:
                return "   ";
            case 4:
                return "     ";
            default:
                final StringBuffer buffer = new StringBuffer(length);
                for (int i = 0; i < length; i++)
                {
                    buffer.append(' ');
                }
                return buffer.toString();
        }
    }

    private static final class Mappings
    {
        private final List leftContent = new ArrayList();
        private final List rightContent = new ArrayList();

        Mappings()
        {
        }

        private static void appendContent(final String str, final List list)
        {
            if (str == null)
            {
                return;
            }
            try
            {
                final BufferedReader reader = new BufferedReader(new StringReader(str));
                String line;
                while ((line = reader.readLine()) != null)
                {
                    list.add(line);
                }
            }
            catch (IOException ignore)
            {
                throw new AssertionError("IOException even while reading string");
            }
        }

        private static void prependContent(final String str, final List list)
        {
            if (str == null)
            {
                return;
            }
            try
            {
                final BufferedReader reader = new BufferedReader(new StringReader(str));
                String line;
                int count = 0;
                while ((line = reader.readLine()) != null)
                {
                    list.add(count, line);
                    count++;
                }
            }
            catch (IOException ignore)
            {
                throw new AssertionError("IOException even while reading string");
            }
        }

        void appendLeftContent(final String str)
        {
            appendContent(str, leftContent);
        }

        void appendRightContent(final String str)
        {
            appendContent(str, rightContent);
        }

        void prependLeftContent(final String str)
        {
            prependContent(str, leftContent);
        }

        void prependRightContent(final String str)
        {
            prependContent(str, rightContent);
        }

        void fillUp()
        {
            if (leftContent.isEmpty())
            {
                leftContent.add("");
            }
            if (rightContent.isEmpty())
            {
                rightContent.add("");
            }
        }

        private static String getContent(final List list)
        {
            final StringBuffer buffer = new StringBuffer(1024);
            for (int i = 0; i < list.size(); i++)
            {
                final String s = (String) list.get(i);
                buffer.append(s);
                buffer.append("\n");
            }
            return buffer.toString();
        }

        int getLeftContentSize()
        {
            return leftContent.size();
        }

        int getRightContentSize()
        {
            return rightContent.size();
        }

        String getLeftContent()
        {
            return getContent(leftContent);
        }

        String getRightContent()
        {
            return getContent(rightContent);
        }

        boolean isEmpty()
        {
            return leftContent.isEmpty() && rightContent.isEmpty();
        }
    }
}
