package org.nkumar.ant.trace.ui;

import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.ScrollingModel;
import com.intellij.openapi.editor.event.VisibleAreaEvent;
import com.intellij.openapi.editor.event.VisibleAreaListener;

/**
 * The sole purpose of this class is to make
 * both sides of the pane scroll in some proportion simultaneously.
 */
final class CommonScroller
{
    private CommonScroller()
    {
    }

    /**
     * Given a vertical offset in one pane what offset does it map to in the other pane.
     * This depends on the data and the mapping of lines.
     */
    private static int mapOffset(final UiDataHolder dataHolder, final int offset, final boolean leftToRight)
    {
        final int[] fromLineInterval =
                leftToRight ? dataHolder.getLeftLineIntervals() : dataHolder.getRightLineIntervals();
        final int[] toLineInterval =
                leftToRight ? dataHolder.getRightLineIntervals() : dataHolder.getLeftLineIntervals();
        final int lineHeight = dataHolder.getLeftEditor().getLineHeight();
        final int fromLine = offset / lineHeight;
        int intervalStartIndex = 0;
        for (; intervalStartIndex < fromLineInterval.length - 1 && fromLine >= fromLineInterval[intervalStartIndex];
                intervalStartIndex++)
        {
        }
        intervalStartIndex--;
        final int fromLineStart = fromLineInterval[intervalStartIndex];
        final int fromLineEnd = fromLineInterval[intervalStartIndex + 1];
        final int toLineStart = toLineInterval[intervalStartIndex];
        final int toLineEnd = toLineInterval[intervalStartIndex + 1];
        return (toLineStart * lineHeight) +
                (offset - fromLineStart * lineHeight) * (toLineEnd - toLineStart) / (fromLineEnd - fromLineStart);
    }

    public static void registerListeners(final UiDataHolder dataHolder)
    {
        final MasterScroller masterScroller = new MasterScroller();
        dataHolder.getLeftEditor().getScrollingModel()
                .addVisibleAreaListener(new CommonScrollerListener(dataHolder, true, masterScroller));
        dataHolder.getRightEditor().getScrollingModel()
                .addVisibleAreaListener(new CommonScrollerListener(dataHolder, false, masterScroller));
    }

    /**
     * Scroll listener.
     * Srolling in one will make the opposite pane to scroll too.
     */
    private static final class CommonScrollerListener implements VisibleAreaListener
    {
        private final UiDataHolder dataHolder;
        private final boolean leftToRight;
        private final MasterScroller masterScroller;

        public CommonScrollerListener(final UiDataHolder dataHolder,
                final boolean leftToRight, final MasterScroller masterScroller)
        {
            this.dataHolder = dataHolder;
            this.leftToRight = leftToRight;
            this.masterScroller = masterScroller;
        }

        public void visibleAreaChanged(final VisibleAreaEvent visibleAreaEvent)
        {
            if (masterScroller.isMaster(this))
            {
                final Editor fromEditor = leftToRight ? dataHolder.getLeftEditor() : dataHolder.getRightEditor();
                final Editor toEditor = leftToRight ? dataHolder.getRightEditor() : dataHolder.getLeftEditor();
                final int thirdMarker = (int) fromEditor.getScrollingModel().getVisibleArea().getHeight() / 3;
                final int fromMarker = thirdMarker + fromEditor.getScrollingModel().getVerticalScrollOffset();
                final int toMarker = mapOffset(dataHolder, fromMarker, leftToRight);
                final ScrollingModel smodel = toEditor.getScrollingModel();
                smodel.disableAnimation();
                smodel.scrollVertically(toMarker - thirdMarker);
                smodel.enableAnimation();
            }
        }
    }

    //no idea whether this is a hack or this is the way UI is developed
    //One scroller when moved tries to move the other scroller,
    //which generates scroll event for that scroller so it, in turn,
    //tries to scroll the previous pane and this could continue
    //so the pane which starts the scroll is the master scroller
    //till some time after he stops scrolling.
    private static final class MasterScroller
    {
        private static final long GAP = 80;
        private VisibleAreaListener listener;
        private long time;

        boolean isMaster(final VisibleAreaListener lis)
        {
            final long now = System.currentTimeMillis();
            if (now - time < GAP)
            {
                time = now;
                return listener.equals(lis);
            }
            else
            {
                time = now;
                listener = lis;
                return true;
            }
        }
    }
}
