package org.nkumar.ant.trace.ui;

import com.intellij.openapi.editor.event.VisibleAreaEvent;
import com.intellij.openapi.editor.event.VisibleAreaListener;
import com.intellij.openapi.ui.Splitter;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.util.ArrayList;
import java.util.List;

/**
 * This class behaves as the splitter as in JSplitter.
 * This splitter is wide enough and shows links between
 * corresponding sections in the two panes.
 */
final class ConnectingSplitter extends Splitter
{
    private final UiDataHolder dataHolder;

    public ConnectingSplitter(final UiDataHolder dataHolder)
    {
        this.dataHolder = dataHolder;
        final MyVisibleAreaListener areaListener = new MyVisibleAreaListener();
        dataHolder.getLeftEditor().getScrollingModel().addVisibleAreaListener(areaListener);
        dataHolder.getRightEditor().getScrollingModel().addVisibleAreaListener(areaListener);
    }

    protected Divider createDivider()
    {
        return new ConnectingDivider();
    }

    private final class ConnectingDivider extends Splitter.Divider
    {
        public void paint(final Graphics graphics)
        {
            super.paint(graphics);
            //TODO prevent painting of polygons that are not visible
            final Polygon[] polygons = createAllThePolygons(getWidth(),
                    dataHolder.getLeftEditor().getLineHeight(),
                    dataHolder.getLeftLineIntervals(), dataHolder.getRightLineIntervals(),
                    dataHolder.getLeftEditor().getScrollingModel().getVerticalScrollOffset(),
                    dataHolder.getRightEditor().getScrollingModel().getVerticalScrollOffset());
            graphics.setColor(AntTracePanel.MY_GRAY);
            final Graphics2D graphics2D = (Graphics2D) graphics.create(0, 0, getWidth(), getEditorHeight());
            for (int i = 0; i < polygons.length; i++)
            {
                graphics2D.fill(polygons[i]);
            }
        }

        private int getEditorHeight()
        {
            try
            {
                //just suspecting that visible area might be null
                return (int) dataHolder.getLeftEditor().getScrollingModel().getVisibleArea().getHeight();
            }
            catch (Exception ignore)
            {
                return dataHolder.getLeftEditor().getComponent().getHeight();
            }
        }
    }

    /**
     * Create all the polygons to connect the left content with the right.
     * Note that all of them might not be visible finally
     * @param width width of the divider
     * @param lineHeight height of each line in the editors
     * @param leftIntervals line intervals for the left editor
     * @param rightIntervals line intervals for the right editor
     * @param leftOffset pixels by which the left editors has scrolled up
     * @param rightOffset pixels by which the right editors has scrolled up
     * @return array of all the connector polygons
     */
    private static Polygon[] createAllThePolygons(final int width, final int lineHeight,
            final int[] leftIntervals, final int[] rightIntervals, final int leftOffset, final int rightOffset)
    {
        if (leftIntervals.length != rightIntervals.length)
        {
            throw new AssertionError("intervals are not of same length");
        }
        final List list = new ArrayList();
        //create connecting polygons for odd intervals
        for (int i = 1; i < leftIntervals.length - 1; i += 2)
        {
            list.add(new Polygon(new int[]{0, width, width, 0},
                    new int[]{leftIntervals[i] * lineHeight - leftOffset,
                            rightIntervals[i] * lineHeight - rightOffset,
                            rightIntervals[i + 1] * lineHeight - rightOffset,
                            leftIntervals[i + 1] * lineHeight - leftOffset},
                    4));
        }
        return (Polygon[]) list.toArray(new Polygon[list.size()]);
    }

    private final class MyVisibleAreaListener implements VisibleAreaListener
    {
        public void visibleAreaChanged(final VisibleAreaEvent visibleAreaEvent)
        {
            getDivider().repaint();
        }
    }
}
